@@ -1,3 +1,6 @@
+from __future__ import absolute_import
+from __future__ import division
+from __future__ import print_function
 import sys
 
 PY2 = sys.version_info[0] < 3
